% ace_ase.m
%
% This script follows the implementation of "neuronlike adaptive
% elements" described in [1]. It relies on 4 external files: f.m,
% noise.m, state_dec.m, and cart_pole.m. See each file header for
% details.
%
% NOTES: - Currently, cart_pole.m simulates dynamics according to
%          a model slightly different from that given in [1]. See
%          the cart-pole-dynamics page on the MIL Wiki website
%          for appropriate citation.
%
% Scott Livingston <slivingston@caltech.edu>
% 31 Dec 2008
% Machine Intelligence Lab, EECS Dept., Univ. of Tennessee, Knoxville
%
% Released under the GNU General Public License, version 3, see
% GPLv3.txt or README for details.

% options:
visualize = false;
verbose = false;

% parameters (as listed in [1]):
gamma = .95; %reward discount
alpha = 1000; %ASE step-size
beta = 0.5; %ACE step-size
delta = 0.9; %ASE trace decay
lambda = 0.8; %ACE trace decay
mu = 0; %action noise mean (Gaussian)
sigma = 0.01; %action noise standard deviation (Gaussian)

% number of possible state indices (i.e., state index can take a
% value from 1 through state_ind_len.
state_ind_len = 162;

action_set = [-10 10]; %assumed to only have two possible actions

if ~exist('w','var') || ~exist('v','var')
    w = zeros(state_ind_len,1);
    v = zeros(state_ind_len,1);
end

% bounds:
max_ang = pi/15; % rad
max_x = 2.4; % m
min_len = 20e3; % minimum number of time steps to consider episode successful
max_len = 20e3; % maximum episode length before early termination (without failure)
max_eps = 1e3; % maximum number of allowed training episodes to achieve minimum goal duration (min_len)

% environment parameters (these MUST match the corresponding values in cart_pole.m)
pole_length = 0.5*2; % full pole length (not just half as in cart_pole.m)
tau = 0.02; % time between each step (in s)

if visualize
    axis_range = [-3 3 0 2]; % axis range of graph to be used by cpvisual();
                             % to reduce shape distortion, this range should be
                             % square, or the figure sent to cpvisual() should
                             % be prepared in a similar manner.
    hf = figure; % generate a new figure to be used by cpvisual()

    tic; % initialize timer, which is used to make animation with cpvisual()
         % more realistic.
end


% history of episode lengths
BUFFER_SIZE = 10e3;
if ~exist('ep_count','var')
    ep_hist = zeros(1,BUFFER_SIZE);
    ep_count = 0;
end

ep_len = 0; % allow entry into main loop

% learning loop
while ep_len < min_len

    if ep_count+1 > max_eps %maximum number of training episodes reached?
       fprintf( 'Fail!\n' );
       break;
    end

    ep_len = 0;

    state = [0 0 0 0]'; % initial state (note, it is a column vector)

    % zero the traces
    e_trace = zeros(size(w));
    x_trace = zeros(size(v));

    % select initial action and determine p_prev (evaluation
    % of initial state, in this case)
    state_ind = state_dec(state);
    action = f( w(state_ind) + noise(mu,sigma) );
    p_prev = v(state_ind);

    % episode loop
    while abs(state(3)) < max_ang && abs(state(1)) < max_x && ep_len < max_len

        % take selected action
        next_state = cart_pole( state, action_set((action>0)+1) );
        next_state_ind = state_dec(next_state);

        % determine reward
        if abs(next_state(3)) > max_ang || abs(next_state(1)) > max_x
            % terminal state is a special case
            r = -1;
            p = 0; %no value estimate for terminal state
        else
            r = 0;
            p = v(next_state_ind); %current value estimate
        end

        % internal reinforcement
        r_pred = r + gamma*p - p_prev;

        % ASE update
        e_trace = delta*e_trace;
        e_trace(state_ind) = e_trace(state_ind) + (1-delta)*action;
        w = w + alpha*r_pred*e_trace;

        % ACE update
        x_trace = lambda*x_trace;
        x_trace(state_ind) = x_trace(state_ind) + (1-lambda);
        v = v + beta*r_pred*x_trace;

        if visualize
            % graph current state:
            while toc < tau, end % wait until at least the time between state
                             % changes, tau, has passed since the last graph.
            cpvisual( hf, pole_length, state, axis_range );
            tic; % reset timer (used for realistic visualization)
        end

        % determine next action if failure did not occur
        if abs(next_state(3)) < max_ang && abs(next_state(1)) < max_x
            action = f( w(next_state_ind) + noise(mu,sigma) );
        end

        % update world
        state = next_state; 
        state_ind = state_dec(state);
        p_prev = p;

        ep_len = ep_len + 1;
    
    end

    ep_count = ep_count+1; %increment number of episodes
    ep_hist(ep_count) = ep_len;
    if rem(ep_count,BUFFER_SIZE) == 0
        tmp_ep_hist = ep_hist;
        ep_hist = zeros(1,ep_count+BUFFER_SIZE);
        ep_hist(1:ep_count) = tmp_ep_hist;
        clear tmp_ep_hist;
    end

    if verbose
        home
        fprintf( 'Episode #%d length = %d\n', ep_count, ep_len );
        fflush( stdout ); %Octave requires this to avoid buffering.
    end

end

fprintf( 'Episode #%d length = %d\n', ep_count, ep_len );
