% example.m - a simple script which demonstrates usage of several tools:
%   cart_pole.m - the cart-pole dynamics function
%   cpvisual.m  - a simple system visualizer
%   cptool.m    - a GUI-based tool for analyzing episode data.
%
%   See their (in-file) help information for usage and operation details.
%   This script will generate an episode using random actions;
%   specifically, at each state the agent applies a random force,
%   uniformly distributed from -1 through 1 Newtons, to the cart. The
%   results are saved to an ASCII text file called "example_data.txt", and
%   the state of each time step is graphed on-line using cpvisual().
%
%   After example.m is executed, use "cptool( 'example_data.txt' )" to open
%   the episode data in cptool for analysis. Also, if cptool is already
%   in operation, use the menu File->Open to load example_data.txt .
%
%
% Written by Scott Livingston, October 2007.
%
% Machine Intelligence Lab, EECS - Univ. of Tennessee
%
% Released under the GNU General Public License, version 3, see
% GPLv3.txt or README for details.


% output episode results
outfile = fopen('example_data.txt', 'wt');

% environment parameters (these MUST match the corresponding values in cart_pole.m)
pole_length = 0.5*2; % full pole length (not just half as in cart_pole.m)
tau = 0.02; % time between each step (in s)

axis_range = [-3 3 0 2]; % axis range of graph to be used by cpvisual();
                         % to reduce shape distortion, this range should be
                         % square, or the figure sent to cpvisual() should
                         % be prepared in a similar manner.
hf = figure; % generate a new figure to be used by cpvisual()

state = [0 0 0 0]'; % initial state (note, it is a column vector)

tic; % initialize timer, which is used to make animation with cpvisual()
     % more realistic.

% the main loop; each iteration is a step in the episode
while state(3) > -pi/2 && state(3) < pi/2 % while pole is not horizontal
    
    % action selection:
    action = rand*2-1; % random force between -1 and 1
    
    % output step information to episode data file
    fprintf( outfile, '%f %f %f %f %f\n', state', action );
    
    % graph current state:
    while toc < tau, end % wait until at least the time between state
                         % changes, tau, has passed since the last graph.
    cpvisual( hf, pole_length, state, axis_range );
    tic; % reset timer (used for realistic visualization)
    
    % take action; detect new state:
    state = cart_pole( state, action );
    
    % at this point, some sort of reinforcement learning backup operation
    % would be performed; other components of such an algorithm might be
    % placed here as well.
    
end

fclose( outfile ); % close the episode data file
